# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from typing import List, Dict


class Vertex:

    __slots__ = [r'_id', r'_in', r'_out']

    def __init__(self, id_: str):

        self._id = id_

        self._in = {}
        self._out = {}

    def __repr__(self):
        return f'<Vertex: {self._id}>'

    @property
    def id(self):
        return self._id


class Edge:

    __slots__ = [r'_in', r'_out']

    def __init__(self, source: Vertex, target: Vertex):

        if source._id == target._id:
            raise Exception(f'vertices like same: {source._id} => {target._id}')

        # 父类
        self._in = source

        # 子类
        self._out = target


class Tree:

    def __init__(self, vertex: Vertex):

        self._steps = [[vertex]]
        self._nodes = {}

    @property
    def steps(self) -> List[List[str]]:

        return self._steps

    @property
    def nodes(self) -> Dict[str, List[str]]:

        return self._nodes

    def find_in(self) -> Dict[str, List[str]]:

        nodes = {}
        vertices = []

        for vertex in self._steps[-1]:

            _vertices = [v._in for v in vertex._in.values() if v._in._id not in self._nodes]

            if _vertices:
                vertices.extend(_vertices)
                nodes[vertex._id] = [v._id for v in _vertices]

        if nodes and vertices:
            self._nodes.update(nodes)
            self._steps.append(vertices)

        return nodes

    def find_out(self) -> Dict[str, List[str]]:

        nodes = {}
        vertices = []

        for vertex in self._steps[-1]:

            _vertices = [v._out for v in vertex._out.values() if v._out._id not in self._nodes]

            if _vertices:
                vertices.extend(_vertices)
                nodes[vertex._id] = [v._id for v in _vertices]

        if nodes and vertices:
            self._nodes.update(nodes)
            self._steps.append(vertices)

        return nodes





# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from typing import List, Dict

from .entity import Vertex, Edge, Tree


class Graph:

    def __init__(self):

        self._vertices = {}

    def __len__(self):

        return len(self._vertices)

    def clear(self):

        for vertex in self._vertices.values():
            vertex._in.clear()
            vertex._out.clear()

        self._vertices.clear()

    def find_vertex(self, id_: str) -> Vertex:

        return self._vertices.get(id_, None)

    def find_vertices(self, ids: List[str]) -> Dict[str, Vertex]:

        return {id_: self._vertices[id_] for id_ in ids if id_ in self._vertices}

    def add_vertex(self, id_: str) -> Vertex:

        if id_ in self._vertices:
            vertex = self._vertices[id_]
        else:
            vertex = self._vertices[id_] = Vertex(id_)

        return vertex

    def add_edge(self, source_id: str, target_id: str) -> Edge:

        source = self.find_vertex(source_id)
        target = self.find_vertex(target_id)

        if source is None or target is None:
            return None

        if source._id == target._id:
            return None

        edge = Edge(source, target)

        # 添加父类
        target._in[source._id] = edge

        # 添加子类
        source._out[target._id] = edge

        return edge

    def del_edge(self, source_id: str, target_id: str):

        # 删除子类
        source = self.find_vertex(source_id)

        if source and target_id in source._out:

            del source._out[target_id]

            if not source._in and not source._out:
                del self._vertices[source_id]

        # 删除父类
        target = self.find_vertex(target_id)

        if target and source_id in target._in:

            del target._in[source_id]

            if not source._in and not source._out:
                del self._vertices[target_id]

    def find_in_edges(self, id_: str) -> List[Edge]:

        vertex = self.find_vertex(id_)

        if vertex is None:
            return None

        return list(vertex._in.values())

    def find_out_edges(self, id_: str) -> List[Edge]:

        vertex = self.find_vertex(id_)

        if vertex is None:
            return None

        return list(vertex._out.values())

    def tree(self, id_: str) -> Tree:

        vertex = self.find_vertex(id_)

        if vertex is None:
            return None

        return Tree(vertex)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import sys

os.chdir(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(r'../'))

from hagworm.extend.asyncio.base import Utils, AsyncCirculatorForSecond
from hagworm.extend.asyncio.command import Launcher, MainProcessAbstract, ChildProcessAbstract


class MainProcess(MainProcessAbstract):

    async def _handle_message(self, message):

        await self._push_server.send(message)


class ChildProcess(ChildProcessAbstract):

    async def _run(self):

        async for idx in AsyncCirculatorForSecond(max_times=5):

            await self._push_client.send(f'{self._process_id}_{idx}')

            message = await self._pull_client.recv()

            Utils.log.info(message)


if __name__ == r'__main__':

    Launcher(daemon=MainProcess(2)).run(ChildProcess())



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from asyncio import Queue
from contextlib import contextmanager, asynccontextmanager

from hagworm.extend.base import Utils


class ObjectPool:
    """对象池实现
    """

    def __init__(self, maxsize):

        self._queue = Queue(maxsize=maxsize)

        for _ in range(maxsize):
            self._queue.put_nowait(self._create_obj())

        Utils.log.info(f'ObjectPool {type(self)} initialized: {self._queue.qsize()}')

    def _create_obj(self):

        raise NotImplementedError()

    @asynccontextmanager
    async def get(self):

        Utils.log.debug(f'ObjectPool {type(self)} size: {self._queue.qsize()}')

        obj = await self._queue.get()

        try:
            yield obj
        except Exception as err:
            raise err
        finally:
            self._queue.put_nowait(obj)

    @contextmanager
    def get_nowait(self):

        Utils.log.debug(f'ObjectPool {type(self)} size: {self._queue.qsize()}')

        obj = self._queue.get_nowait()

        try:
            yield obj
        except Exception as err:
            raise err
        finally:
            self._queue.put_nowait(obj)





# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import time
import numpy

from ntplib import NTPClient

from .base import Utils, MultiTasks
from .task import IntervalTask
from .future import ThreadPool

from ..error import NTPCalibrateError
from ..interface import TaskInterface


class _Interface(TaskInterface):
    """NTP客户端接口定义
    """

    def start(self):
        raise NotImplementedError()

    def stop(self):
        raise NotImplementedError()

    def is_running(self):
        raise NotImplementedError()

    def calibrate_offset(self):
        raise NotImplementedError()

    @property
    def offset(self):
        raise NotImplementedError()

    @property
    def timestamp(self):
        raise NotImplementedError()


class AsyncNTPClient(_Interface):
    """异步NTP客户端类
    """

    @classmethod
    async def create(cls, host):

        client = cls(host)

        await client.calibrate_offset()

        client.start()

        return client

    def __init__(self, host, *, version=2, port=r'ntp', timeout=5, interval=3600, sampling=5):

        self._settings = {
            r'host': host,
            r'version': version,
            r'port': port,
            r'timeout': timeout,
        }

        self._client = NTPClient()
        self._offset = 0

        self._thread_pool = ThreadPool(1)

        self._sync_task = IntervalTask.create(interval, self.calibrate_offset)

        self._sampling = sampling

    def start(self):

        return self._sync_task.start()

    def stop(self):

        return self._sync_task.stop()

    def is_running(self):

        return self._sync_task.is_running()

    async def calibrate_offset(self):

        return await self._thread_pool.run(self._calibrate_offset)

    def _calibrate_offset(self):

        samples = []
        host_name = self._settings[r'host']

        # 多次采样取中位数，减少抖动影响
        for _ in range(self._sampling):
            try:
                resp = self._client.request(**self._settings)
                samples.append(resp.offset)
            except Exception as err:
                Utils.log.warning(f'NTP server {host_name} request error: {err}')

        if samples:
            self._offset = float(numpy.median(samples))
            Utils.log.debug(f'NTP server {host_name} offset median {self._offset} samples: {samples}')
        else:
            raise NTPCalibrateError(f'NTP server {host_name} not available, timestamp uncalibrated')

    @property
    def offset(self):

        return self._offset

    @property
    def timestamp(self):

        return time.time() + self._offset


class AsyncNTPClientPool(_Interface):
    """异步NTP客户端池，多节点取中位数实现高可用
    """

    @classmethod
    async def create(cls, hosts):

        client_pool = cls()

        for host in hosts:
            client_pool.append(host)

        await client_pool.calibrate_offset()

        client_pool.start()

        return client_pool

    def __init__(self):

        self._clients = []
        self._running = False

    def append(self, host, *, version=2, port='ntp', timeout=5, interval=3600, sampling=5):

        client = AsyncNTPClient(host, version=version, port=port, timeout=timeout, interval=interval, sampling=sampling)

        self._clients.append(client)

        if self._running:
            client.start()

    def start(self):

        for client in self._clients:
            client.start()

        self._running = True

    def stop(self):

        for client in self._clients:
            client.stop()

        self._running = False

    def is_running(self):

        return self._running

    async def calibrate_offset(self):

        tasks = MultiTasks()

        for client in self._clients:
            tasks.append(client.calibrate_offset())

        await tasks

    @property
    def offset(self):

        samples = []

        for client in self._clients:
            samples.append(client.offset)

        return float(numpy.median(samples))

    @property
    def timestamp(self):

        return time.time() + self.offset



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

from threading import Thread

from etcd3gw.client import Etcd3Client

from ...extend.config import ConfigureBase
from ...extend.error import catch_error


class Etcd3Configure(ConfigureBase):
    """基于etcd3的配置类
    """

    __slots__ = [r'_client', r'_key_prefix', r'_watch_iterator', r'_watch_cancel', r'_task']

    def __init__(self, host, port=2379, *, key_prefix, api_path=r'/v3/', **kwargs):

        super().__init__()

        self._client = Etcd3Client(host, port, **kwargs, api_path=api_path)

        self._key_prefix = key_prefix
        self._watch_iterator, self._watch_cancel = self._client.watch_prefix(self._key_prefix)

        self._task = Thread(target=self._do_task)
        self._task.start()

    def _do_task(self):

        for _item in self._watch_iterator:

            with catch_error():

                if r'value' not in _item[r'kv']:
                    continue

                _key_section = _item[r'kv'][r'key'].decode()

                if _key_section.find(self._key_prefix) == 0:
                    _key_section = _key_section.replace(self._key_prefix, r'', 1)

                if _key_section not in self._key_section:
                    continue

                _key, _section, _type = self._key_section.get(_key_section)

                if _type in (str, int, float, bool):
                    self.__setattr__(_key, _type(_item[r'kv'][r'value'].decode()))
                else:
                    self.__setattr__(_key, _type.decode(_item[r'kv'][r'value'].decode()))

    def close(self):

        self._watch_cancel()
        self._task.join(1)

    def load(self):

        setting = {}

        _resp = self._client.get_prefix(self._key_prefix)

        for _item in _resp:

            _key = _item[1][r'key'].decode()

            if _key.find(self._key_prefix) == 0:
                _key = _key.replace(self._key_prefix, r'', 1)

            setting[_key] = _item[0].decode()

        for _key, _section, _type in self._key_section.values():

            if _type in (str, int, float, bool):
                self.__setattr__(_key, _type(setting[f'{_section}/{_key}']))
            else:
                self.__setattr__(_key, _type.decode(setting[f'{_section}/{_key}']))

        return self



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import aio_pika


class RabbitMQConsumer(aio_pika.RobustConnection):
    """RabbitMQ消费者
    """

    def __init__(self, url, **kwargs):

        super().__init__(url, **kwargs)

        self._channel = None
        self._queue = None

    @property
    def current_channel(self):

        return self._channel

    @property
    def current_queue(self):

        return self._queue

    async def initialize(self, queue_name, consume_qos=1, *, queue_config=None):

        await self.connect()
        await self.ready()

        self._channel = await self.channel()
        await self._channel.set_qos(prefetch_count=consume_qos)

        if queue_config is None:
            self._queue = await self._channel.declare_queue(queue_name)
        else:
            self._queue = await self._channel.declare_queue(queue_name, **queue_config)

    async def release(self):

        await self._channel.close()
        await super().close()

    async def get(self, *, no_ack=False, timeout=1):

        await self._queue.get(no_ack=no_ack, timeout=timeout)

    async def consume(self, consume_func, *, no_ack=False):

        await self._queue.consume(consume_func, no_ack=no_ack)


class RabbitMQConsumerForExchange(RabbitMQConsumer):
    """RabbitMQ注册到交换机的消费者(默认生成排它队列注册)
    """

    def __init__(self, url, **kwargs):
        
        super().__init__(url, **kwargs)

        self._exchange = None

    @property
    def current_exchange(self):

        return self._exchange

    async def initialize(self, exchange_name, consume_qos=1, *, queue_config=None, routing_key=None):

        await self.connect()
        await self.ready()

        self._channel = await self.channel()
        await self._channel.set_qos(prefetch_count=consume_qos)

        if queue_config is None:
            self._queue = await self._channel.declare_queue(exclusive=True)
        else:
            self._queue = await self._channel.declare_queue(**queue_config)

        self._exchange = await self._channel.get_exchange(exchange_name)

        await self._queue.bind(self._exchange, routing_key)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import os
import ssl
import aiohttp

from contextlib import contextmanager

from ..struct import IntEnum
from ..interface import ContextManager

from .base import Utils, AsyncCirculatorForSecond
from .buffer import FileBuffer


class STATE(IntEnum):

    PENDING = 0x00
    FETCHING = 0x01
    SUCCESS = 0x02
    FAILURE = 0x03


DEFAULT_TIMEOUT = aiohttp.client.ClientTimeout(total=60, connect=10, sock_read=60, sock_connect=10)
DOWNLOAD_TIMEOUT = aiohttp.client.ClientTimeout(total=600, connect=10, sock_read=600, sock_connect=10)

DEFAULT_SSL_CONTEXT = ssl.create_default_context(
    cafile=os.path.join(
        os.path.split(os.path.abspath(__file__))[0],
        r'../../static/cacert.pem'
    )
)


@contextmanager
def _catch_error():
    """异常捕获

    通过with语句捕获异常，代码更清晰，还可以使用Ignore异常安全的跳出with代码块

    """

    try:

        yield

    except aiohttp.ClientResponseError as err:

        if err.status < 500:
            Utils.log.warning(err)
        else:
            Utils.log.error(err)

    except Exception as err:

        Utils.log.error(err)


class _AsyncCirculator(AsyncCirculatorForSecond):

    async def _sleep(self):

        await Utils.sleep(self._current)


def _json_decoder(val, **kwargs):

    try:
        return Utils.json_decode(val, **kwargs)
    except Exception as err:
        Utils.log.warning(f'http client json decode error: {err} => {val}')


def create_timeout(total=None, connect=None, sock_read=None, sock_connect=None) -> aiohttp.client.ClientTimeout:
    """生成超时配置对象

    Args:
        total: 总超时时间
        connect: 从连接池中等待获取连接的超时时间
        sock_read: Socket数据接收的超时时间
        sock_connect: Socket连接的超时时间

    """

    return aiohttp.client.ClientTimeout(
        total=total, connect=connect,
        sock_read=sock_read, sock_connect=sock_connect
    )


def create_ssl_context(cafile, *, certfile=None, keyfile=..., password=...) -> ssl.SSLContext:

    ssl_context = ssl.create_default_context(cafile=cafile)

    if certfile is not None:
        ssl_context.load_cert_chain(certfile, keyfile, password)

    return ssl_context


class Result(dict):

    def __init__(self, status, headers, cookies, body):

        super().__init__(status=status, headers=headers, cookies=cookies, body=body)

    def __bool__(self):

        return (self.status >= 200) and (self.status <= 299)

    @property
    def status(self):

        return self.get(r'status')

    @property
    def headers(self):

        return self.get(r'headers')

    @property
    def cookies(self):

        return self.get(r'cookies')

    @property
    def body(self):

        return self.get(r'body')


class _HTTPClient:
    """HTTP客户端基类
    """

    def __init__(self, retry_count=5, timeout=DEFAULT_TIMEOUT, *, ssl_context=DEFAULT_SSL_CONTEXT, **kwargs):

        self._ssl_context = ssl_context
        self._retry_count = retry_count

        self._session_config = kwargs
        self._session_config[r'timeout'] = timeout
        self._session_config.setdefault(r'raise_for_status', True)

    def _encode_body(self, headers, data):

        if headers is None:
            headers = {}

        if isinstance(data, dict):
            headers.setdefault(r'Content-Type', r'application/x-www-form-urlencoded')

        return headers, data

    async def _handle_response(self, response):

        return await response.read()

    def create_form_data(self) -> aiohttp.FormData:

        return aiohttp.FormData()

    async def send_request(
            self, method, url, data=None, params=None, cookies=None, headers=None, **settings
    ) -> Result:

        response = None

        headers, data = self._encode_body(headers, data)

        settings[r'data'] = data
        settings[r'params'] = params
        settings[r'cookies'] = cookies
        settings[r'headers'] = headers

        Utils.log.debug(
            r'{0} {1} => {2}'.format(
                method,
                url,
                str({key: val for key, val in settings.items() if isinstance(val, (str, list, dict))})
            )
        )

        settings.setdefault(r'ssl', self._ssl_context)

        async for times in _AsyncCirculator(max_times=self._retry_count):

            try:

                async with aiohttp.ClientSession(**self._session_config) as _session:

                    async with _session.request(method, url, **settings) as _response:

                        response = Result(
                            _response.status,
                            dict(_response.headers),
                            {key: val.value for key, val in _response.cookies.items()},
                            await self._handle_response(_response)
                        )

            except aiohttp.ClientResponseError as err:

                # 重新尝试的话，会记录异常，否则会继续抛出异常

                if err.status < 500:
                    raise err
                elif times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except aiohttp.ClientError as err:

                if times >= self._retry_count:
                    raise err
                else:
                    Utils.log.warning(err)
                    continue

            except Exception as err:

                raise err

            else:

                Utils.log.info(f'{method} {url} => status:{response.status}')
                break

            finally:

                if times > 1:
                    Utils.log.warning(f'{method} {url} => retry:{times}')

        return response


class _HTTPTextMixin:
    """Text模式混入类
    """

    async def _handle_response(self, response):

        return await response.text()


class _HTTPJsonMixin:
    """Json模式混入类
    """

    async def _handle_response(self, response):

        return await response.json(encoding=r'utf-8', loads=_json_decoder, content_type=None)


class _HTTPJsonRequestMixin(_HTTPJsonMixin):
    """Json请求模式混入类
    """

    def _encode_body(self, headers, data):

        if headers is None:
            headers = {}

        if isinstance(data, dict):
            headers.setdefault(r'Content-Type', r'application/json')
            data = Utils.json_encode(data)

        return headers, data


class _HTTPTouchMixin:
    """Touch模式混入类，不接收body数据
    """

    async def _handle_response(self, response):

        return dict(response.headers)


class HTTPClient(_HTTPClient):
    """HTTP客户端，普通模式
    """

    async def get(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_GET, url, None, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def options(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_OPTIONS, url, None, params, cookies=cookies, headers=headers)

            result = resp.headers

        return result

    async def head(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_HEAD, url, None, params, cookies=cookies, headers=headers)

            result = resp.headers

        return result

    async def post(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_POST, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def put(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_PUT, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def patch(self, url, data=None, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_PATCH, url, data, params, cookies=cookies, headers=headers)

            result = resp.body

        return result

    async def delete(self, url, params=None, *, cookies=None, headers=None):

        result = None

        with _catch_error():

            resp = await self.send_request(aiohttp.hdrs.METH_DELETE, url, None, params, cookies=cookies, headers=headers)

            result = resp.body

        return result


class HTTPTextClient(_HTTPTextMixin, HTTPClient):
    """HTTP客户端，Text模式
    """
    pass


class HTTPJsonClient(_HTTPJsonMixin, HTTPClient):
    """HTTP客户端，Json模式
    """
    pass


class HTTPJsonRequestClient(_HTTPJsonRequestMixin, HTTPClient):
    """HTTP客户端，Json请求模式
    """
    pass


class HTTPTouchClient(_HTTPTouchMixin, HTTPClient):
    """HTTP客户端，Touch模式
    """
    pass


class HTTPClientPool(HTTPClient):
    """HTTP带连接池客户端，普通模式
    """

    def __init__(self,
                 retry_count=5, use_dns_cache=True, ttl_dns_cache=10,
                 limit=100, limit_per_host=0, timeout=DEFAULT_TIMEOUT,
                 **kwargs
                 ):

        super().__init__(retry_count, timeout, **kwargs)

        self._tcp_connector = aiohttp.TCPConnector(
            use_dns_cache=use_dns_cache,
            ttl_dns_cache=ttl_dns_cache,
            ssl=self._ssl_context,
            limit=limit,
            limit_per_host=limit_per_host,
        )

        self._session_config[r'connector'] = self._tcp_connector
        self._session_config[r'connector_owner'] = False

    async def close(self):

        if not self._tcp_connector.closed:
            await self._tcp_connector.close()


class HTTPTextClientPool(_HTTPTextMixin, HTTPClientPool):
    """HTTP带连接池客户端，Text模式
    """
    pass


class HTTPJsonClientPool(_HTTPJsonMixin, HTTPClientPool):
    """HTTP带连接池客户端，Json模式
    """
    pass


class HTTPJsonRequestClientPool(_HTTPJsonRequestMixin, HTTPClientPool):
    """HTTP带连接池客户端，Json请求模式
    """
    pass


class HTTPTouchClientPool(_HTTPTouchMixin, HTTPClientPool):
    """HTTP带连接池客户端，Touch模式
    """
    pass


class Downloader(_HTTPClient):
    """HTTP文件下载器
    """

    def __init__(self, file, retry_count=5, timeout=DOWNLOAD_TIMEOUT, **kwargs):

        super().__init__(retry_count, timeout, **kwargs)

        self._file = file

        self._state = STATE.PENDING

        self._response = None

    @property
    def file(self):

        return self._file

    @property
    def state(self):

        return self._state

    @property
    def finished(self):

        return self._state in (STATE.SUCCESS, STATE.FAILURE)

    @property
    def response(self):

        return self._response

    async def _handle_response(self, response):

        if self._state != STATE.PENDING:
            return

        self._state = STATE.FETCHING
        self._response = response

        with open(self._file, r'wb') as stream:

            try:

                while True:

                    chunk = await response.content.read(65536)

                    if chunk:
                        stream.write(chunk)
                    else:
                        break

            except Exception as err:

                Utils.log.error(err)

                self._state = STATE.FAILURE

            else:

                self._state = STATE.SUCCESS

        if self._state != STATE.SUCCESS and os.path.exists(self._file):
            os.remove(self._file)

    async def fetch(self, url, *, params=None, cookies=None, headers=None):

        result = False

        try:

            await self.send_request(aiohttp.hdrs.METH_GET, url, None, params, cookies=cookies, headers=headers)

            result = (self._state == STATE.SUCCESS)

        except Exception as err:

            Utils.log.error(err)

            self._state = STATE.FAILURE

        return result


class DownloadBuffer(ContextManager, Downloader):
    """HTTP文件下载器(临时文件版)
    """

    def __init__(self, timeout=DOWNLOAD_TIMEOUT, **kwargs):

        super().__init__(FileBuffer(), 1, timeout, **kwargs)

    def _context_release(self):

        self.close()

    def close(self):

        self._file.close()

    async def _handle_response(self, response):

        if self._state != STATE.PENDING:
            return

        self._state = STATE.FETCHING
        self._response = response

        try:

            while True:

                chunk = await response.content.read(65536)

                if chunk:
                    self._file.write(chunk)
                else:
                    break

        except Exception as err:

            Utils.log.error(err)

            self._state = STATE.FAILURE

        else:

            self._state = STATE.SUCCESS



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import dataclasses

from configparser import RawConfigParser

from .base import Utils
from .event import EventDispatcher


class HostType(str):

    @classmethod
    def decode(cls, val):

        if not val:
            return None

        host, port = val.split(r':', 2)
        return host.strip(), int(port.strip())


class JsonType(str):

    @classmethod
    def decode(cls, val):

        if not val:
            return None

        return Utils.json_decode(val)


class StrListType(str):

    @classmethod
    def decode(cls, val):
        return Utils.split_str(val)


class IntListType(str):

    @classmethod
    def decode(cls, val):
        return Utils.split_int(val)


class FloatListType(str):

    @classmethod
    def decode(cls, val):
        return Utils.split_float(val)


class Field:

    __slots__ = [r'section']

    def __init__(self, section):

        self.section = section


class ConfigureMetaclass(type):
    """配置类元类，增加dataclass修饰
    """

    def __new__(mcs, name, bases, attrs):
        return dataclasses.dataclass(init=False)(
            type.__new__(mcs, name, bases, attrs)
        )


class ConfigureBase(metaclass=ConfigureMetaclass):
    """配置类
    """

    __slots__ = [r'_event_dispatcher', r'_key_section']

    def __init__(self):

        super().__init__()

        temp = self.__dataclass_fields__

        self._event_dispatcher = EventDispatcher()

        self._key_section = {
            f'{_field.default.section}/{_key}': (_key, _field.default.section, _field.type,)
            for _key, _field in self.__dataclass_fields__.items()
        }

    def __setattr__(self, key, value):

        super().__setattr__(key, value)

        if key in self.__dataclass_fields__:
            self._event_dispatcher.dispatch(key, key, value)

    def add_listener(self, key, _callable):
        return self._event_dispatcher.add_listener(key, _callable)

    def remove_listener(self, key, _callable):
        return self._event_dispatcher.remove_listener(key, _callable)


class Configure(ConfigureBase):
    """配置类
    """

    __slots__ = [r'_parser']

    def __init__(self):

        super().__init__()

        self._parser = RawConfigParser()

    def _load_options(self):

        for _key, _section, _type in self._key_section.values():

            if _type is str:
                self.__setattr__(_key, self._parser.get(_section, _key))
            elif _type is int:
                self.__setattr__(_key, self._parser.getint(_section, _key))
            elif _type is float:
                self.__setattr__(_key, self._parser.getfloat(_section, _key))
            elif _type is bool:
                self.__setattr__(_key, self._parser.getboolean(_section, _key))
            else:
                self.__setattr__(_key, _type.decode(self._parser.get(_section, _key)))

    def get_option(self, section, option):

        return self._parser.get(section, option)

    def get_options(self, section):

        parser = self._parser

        options = {}

        for option in parser.options(section):
            options[option] = parser.get(section, option)

        return options

    def set_options(self, section, **options):

        if not self._parser.has_section(section):
            self._parser.add_section(section)

        for option, value in options.items():
            self._parser.set(section, option, value)

        self._load_options()

    def read(self, files):

        self._parser.clear()
        self._parser.read(files, r'utf-8')

        self._load_options()

    def read_str(self, val):

        self._parser.clear()
        self._parser.read_string(val)

        self._load_options()

    def read_dict(self, val):

        self._parser.clear()
        self._parser.read_dict(val)

        self._load_options()



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import aio_pika


class RabbitMQProducer:
    """RabbitMQ通用发布者
    """

    def __init__(self, url, *, pool_size=10, connection_config=None):

        self._mq_url = url

        self._connection_config = connection_config if connection_config is not None else {}

        self._connection_pool = aio_pika.pool.Pool(self._create_connection, max_size=pool_size)
        self._channel_pool = aio_pika.pool.Pool(self._create_channel, max_size=pool_size)

    async def _create_connection(self):

        return await aio_pika.connect_robust(self._mq_url, **self._connection_config)

    async def _create_channel(self):

        async with self._connection_pool.acquire() as connection:
            return await connection.channel()

    async def release(self):

        await self._channel_pool.close()
        await self._connection_pool.close()

    def acquire_connection(self):

        return self._connection_pool.acquire()

    def acquire_channel(self):

        return self._channel_pool.acquire()

    async def publish(self, message, routing_key, **kwargs):

        async with self.acquire_channel() as channel:
            await channel.default_exchange.publish(aio_pika.Message(message), routing_key, **kwargs)

    async def batch_publish(self, messages, routing_key, **kwargs):

        async with self.acquire_channel() as channel:
            for message in messages:
                await channel.default_exchange.publish(aio_pika.Message(message), routing_key, **kwargs)


class RabbitMQProducerForExchange(RabbitMQProducer):
    """RabbitMQ使用交换机发布
    """

    def __init__(
            self,
            url, exchange_name, *, pool_size=10, connection_config=None,
            exchange_type=aio_pika.ExchangeType.FANOUT, exchange_config=None
    ):

        super().__init__(url, pool_size=pool_size, connection_config=connection_config)

        self._exchange_name = exchange_name
        self._exchange_type = exchange_type

        self._exchange_config = exchange_config if exchange_config is not None else {}

        self._exchange_pool = aio_pika.pool.Pool(self._create_exchange, max_size=pool_size)

    async def _create_exchange(self):

        async with self.acquire_channel() as channel:
            return await channel.declare_exchange(self._exchange_name, self._exchange_type, **self._exchange_config)

    def acquire_exchange(self):

        return self._exchange_pool.acquire()

    async def publish(self, message, routing_key=r'', **kwargs):

        async with self.acquire_exchange() as exchange:
            await exchange.publish(aio_pika.Message(message), routing_key, **kwargs)

    async def batch_publish(self, messages, routing_key=r'', **kwargs):

        async with self.acquire_exchange() as exchange:
            for message in messages:
                await exchange.publish(aio_pika.Message(message), routing_key, **kwargs)



# -*- coding: utf-8 -*-

__author__ = r'wsb310@gmail.com'

import sys

from terminal_table import Table

from ..extend.interface import FunctorInterface, RunnableInterface

from ..extend.asyncio import command
from ..extend.asyncio.base import Utils, MultiTasks, AsyncCirculator
from ..extend.asyncio.zmq import Subscribe, PublishWithBuffer


SIGNAL_PROTOCOL = r'tcp'
SIGNAL_PORT = 83310
HIGH_WATER_MARK = 0xffffff


class Reporter(Subscribe):

    class _Report:

        def __init__(self):
            self.success = []
            self.failure = []

    def __init__(self):

        global SIGNAL_PROTOCOL, SIGNAL_PORT

        super().__init__(f'{SIGNAL_PROTOCOL}://*:{SIGNAL_PORT}', True)

        self._reports = {}

    async def handle_message(self):

        while True:

            message = await self.recv(True)

            if message is None:
                break

            for name, result, resp_time in message:
                if name and result in (r'success', r'failure'):
                    getattr(self._get_report(name), result).append(resp_time)

    def _get_report(self, name: str) -> _Report:

        if name not in self._reports:
            self._reports[name] = self._Report()

        return self._reports[name]

    def open(self):

        global HIGH_WATER_MARK

        super().open(HIGH_WATER_MARK)

    def get_report_table(self) -> str:

        reports = []

        for key, val in self._reports.items():
            reports.append(
                (
                    key,
                    len(val.success),
                    len(val.failure),
                    r'{:.2%}'.format(len(val.success) / (len(val.success) + len(val.failure))),
                    r'{:.3f}s'.format(sum(val.success) / len(val.success) if len(val.success) > 0 else 0),
                    r'{:.3f}s'.format(min(val.success) if len(val.success) > 0 else 0),
                    r'{:.3f}s'.format(max(val.success) if len(val.success) > 0 else 0),
                )
            )

        return Table.create(
            reports,
            (
                r'EventName',
                r'SuccessTotal',
                r'FailureTotal',
                r'SuccessRatio',
                r'SuccessAveTime',
                r'SuccessMinTime',
                r'SuccessMaxTime',
            ),
            use_ansi=False
        )


class TaskInterface(Utils, RunnableInterface):

    def __init__(self, publisher: PublishWithBuffer):

        self._publisher = publisher

    def success(self, name: str, resp_time: int):

        self._publisher.send(
            (name, r'success', resp_time,)
        )

    def failure(self, name: str, resp_time: int):

        self._publisher.send(
            (name, r'failure', resp_time,)
        )

    async def run(self):
        raise NotImplementedError()


class Runner(Utils, FunctorInterface):

    def __init__(self, task_cls: TaskInterface):

        global SIGNAL_PROTOCOL, SIGNAL_PORT

        self._task_cls = task_cls
        self._publisher = PublishWithBuffer(f'{SIGNAL_PROTOCOL}://localhost:{SIGNAL_PORT}', False)

    async def __call__(self, times, task_num):

        self._publisher.open()

        for _ in range(times):

            tasks = MultiTasks()

            for _ in range(task_num):
                tasks.append(self._task_cls(self._publisher).run())

            await tasks

        await self._publisher.safe_close()


class Launcher(command.Launcher):
    pass


class Daemon(command.Daemon):

    def __init__(self, child_num=2, report_cls=Reporter):

        super().__init__(child_num, False)

        self._report_cls = report_cls

    async def _run(self):

        with self._report_cls() as reporter:

            async for _ in AsyncCirculator():

                self._check_processes()

                if not self.is_active():
                    break

                await reporter.handle_message()

            Utils.log.info(f'\n{reporter.get_report_table()}')

    def __call__(self):

        if self._fork_processes():
            Utils.run_until_complete(self._run())
            sys.exit(0)
